﻿/* WPF 対応チェックボックス付きフォルダ選択コントロール（複数選択可能）

	【プロパティ】
		SelectedPath : ディレクトリの初期値および選択したディレクトリ名（設定しない場合は、ルートディレクトリになる）
    SelectedPathList : チェックを入れた複数のディレクトリのリスト

  【注意】
		このクラスは単独でも使えるが、FolderDialog.xaml とともに使うとダイアログボックスとして使用可能
  
  【著作権】
		本クラスはフリーウエアです。コードの改変・流用は無制限に許可します。ただし、このクラスを使ったことによる
		すべての不都合に対して著作権者は免責されるものとします。

		2012/12/06
		佐藤 正
		e-mail   : pmansato@kanazawa-net.ne.jp
	  HomePage : http://www.kanazawa-net.ne.jp/~pmansato/

 */
using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Runtime.InteropServices;
using System.Windows.Interop; // Imaging
using System.IO; // DriveInfo
using System.ComponentModel;
using System.Diagnostics;

namespace emanual.Wpf.Controls
{
	public partial class FolderBrowserWithCheckBox : UserControl
	{
		private List<DirectoryData> FItemList;
		public static readonly DependencyProperty SelectedPathProperty;
		public static readonly DependencyProperty SelectedPathListProperty;
		public static readonly RoutedEvent SelectedPathChangedEvent;
		public static readonly RoutedEvent SelectedPathListChangedEvent;

		public string SelectedPath // デフォルトは String.Empty
		{
			get { return (string)base.GetValue(SelectedPathProperty); }
			set
			{
				this.OnSelectedPathChanged(new RoutedPropertyChangedEventArgs<string>(this.SelectedPath, value, SelectedPathChangedEvent));
				base.SetValue(SelectedPathProperty, value);
			}
		}

		public List<string> SelectedPathList // デフォルトは空のリスト
		{
			get { return (List<string>)base.GetValue(SelectedPathListProperty); }
			set
			{
				this.OnSelectedPathListChanged(new RoutedPropertyChangedEventArgs<List<string>>(this.SelectedPathList, value, SelectedPathListChangedEvent));
				base.SetValue(SelectedPathListProperty, value);
			}
		}

		// SelectedPath プロパティが変化したときに発生するイベントを設定・取得する
		public event RoutedPropertyChangedEventHandler<string> SelectedPathChanged
		{
			add { this.AddHandler(SelectedPathChangedEvent, value); }
			remove { this.RemoveHandler(SelectedPathChangedEvent, value); }
		}

		// SelectedPathList プロパティが変化したときに発生するイベントを設定・取得する
		public event RoutedPropertyChangedEventHandler<List<string>> SelectedPathListChanged
		{
			add { this.AddHandler(SelectedPathListChangedEvent, value); }
			remove { this.RemoveHandler(SelectedPathListChangedEvent, value); }
		}

		//---------------------------------------------------------------------------------------------
		// static なコンストラクタ
		static FolderBrowserWithCheckBox()
		{
			SelectedPathProperty = DependencyProperty.Register("SelectedPath", typeof(string), typeof(FolderBrowserWithCheckBox),
			 new FrameworkPropertyMetadata(String.Empty));

			SelectedPathChangedEvent = EventManager.RegisterRoutedEvent("SelectedPathChanged",
				RoutingStrategy.Bubble, typeof(RoutedPropertyChangedEventHandler<string>), typeof(FolderBrowserWithCheckBox));

			SelectedPathListProperty = DependencyProperty.Register("SelectedPathList", typeof(List<string>), typeof(FolderBrowserWithCheckBox),
				new FrameworkPropertyMetadata(new List<string>()));

			SelectedPathListChangedEvent = EventManager.RegisterRoutedEvent("SelectedPathsChanged",
				RoutingStrategy.Bubble, typeof(RoutedPropertyChangedEventHandler<List<string>>), typeof(FolderBrowserWithCheckBox));
		}

		//---------------------------------------------------------------------------------------------
		// コンストラクタ
		public FolderBrowserWithCheckBox()
		{
			InitializeComponent();
		}

		//---------------------------------------------------------------------------------------------
		// コントロールをロードするとき、ドライブデータを初期値として設定する
		private void control_Loaded(object sender, RoutedEventArgs e)
		{
			FItemList = new List<DirectoryData>();

			this.SelectedPathList.Clear();

			DriveInfo[] infos = DriveInfo.GetDrives();

			foreach (DriveInfo info in infos)
			{
				var data = new DirectoryData(null, info.Name, this.SelectedPath);
				FItemList.Add(data);
			}

			string dirName = Environment.GetFolderPath(Environment.SpecialFolder.UserProfile);
			var dirData = new DirectoryData(null, dirName, this.SelectedPath);
			FItemList.Add(dirData);

			treeView.ItemsSource = FItemList;

			System.Windows.Input.Keyboard.Focus(treeView);
		}

		//---------------------------------------------------------------------------------------------
		protected virtual void OnSelectedPathChanged(RoutedPropertyChangedEventArgs<string> args)
		{
			this.RaiseEvent(args);
		}

		//---------------------------------------------------------------------------------------------
		protected virtual void OnSelectedPathListChanged(RoutedPropertyChangedEventArgs<List<string>> args)
		{
			this.RaiseEvent(args);
		}

		//---------------------------------------------------------------------------------------------
		private void treeViewItem_Expanded(object sender, RoutedEventArgs e)
		{
			var item = e.Source as TreeViewItem;

			if (item != null)
			{
				item.BringIntoView();

				e.Handled = true;
			}
		}

		//---------------------------------------------------------------------------------------------
		private void treeViewItem_Selected(object sender, RoutedEventArgs e)
		{
			var item = e.Source as TreeViewItem;

			if (item != null)
			{
				var data = item.Header as DirectoryData;

				if (data != null)
				{
					this.SelectedPath = data.FullName;
				}

				e.Handled = true;
			}
		}

		//---------------------------------------------------------------------------------------------
		// チェックボックスにチェックを入れたとき、ディレクトリのフルパスをリストに追加する
		private void chkSelect_Checked(object sender, RoutedEventArgs e)
		{
			var item = e.Source as CheckBox;

			if (item != null)
			{
				this.SelectedPathList.Add((string)item.Content);

				e.Handled = true;
			}
		}

		//---------------------------------------------------------------------------------------------
		// チェックボックスのチェックをはずしたとき、ディレクトリのフルパスのリストから削除する
		private void chkSelect_Unchecked(object sender, RoutedEventArgs e)
		{
			var item = e.Source as CheckBox;

			if (item != null)
			{
				int index = this.SelectedPathList.IndexOf((string)item.Content); // 削除する項目のインデックス
				this.SelectedPathList.RemoveAt(index);
				e.Handled = true;
			}
		}

	} // end of FolderBrowserWithCheckBox class

	//***********************************************************************************************
	// DirectoryData class（ドライブとディレクトリとをあらわすデータ構造を定義するクラス）
	//***********************************************************************************************
	public class DirectoryData : System.ComponentModel.INotifyPropertyChanged
	{
		private DirectoryInfo FDirectoryInfo = null;  // 現在のデータの DirectoryInfo オブジェクト
		private string FDirectoryName;                // ディレクトリの表示名
		private string FFullName;                     // 同、フルパス
		private BitmapSource FDirectoryIcon = null;   // アイコン
		private List<DirectoryData> FSubItems = null; // サブディレクトリのリスト
		private bool FHasSubItems;                    // サブディレクトリを持つとき trye
		private static BitmapSource FDefaultIcon;     // フォルダ用デフォルトのアイコン
		private bool FIsExpanded;                     // 展開するとき true
		private bool FIsSelected;                     // 選択状態にするとき true
		private string FSelectedPath;                 // コントロールの初期値または戻り値

		public event PropertyChangedEventHandler PropertyChanged;

		// プロパティ
		public string DirectoryName { get { return FDirectoryName; } set { FDirectoryName = value; } }
		public string FullName { get { return FFullName; } set { FFullName = value; } }
		public BitmapSource DirectoryIcon { get { return FDirectoryIcon; } set { FDirectoryIcon = value; } }
		public bool IsExpanded { get { return FIsExpanded; } set { FIsExpanded = value; this.OnPropertyChanged("IsExpanded"); } }
		public bool IsSelected { get { return FIsSelected; } set { FIsSelected = value; this.OnPropertyChanged("IsSelected"); } }

		public List<DirectoryData> SubItems
		{
			get
			{
				if (FHasSubItems)
				{
					var list = new List<DirectoryData>();
					var dirInfos = FDirectoryInfo.GetDirectories();

					foreach (var info in dirInfos)
					{
						if ((info.Attributes & FileAttributes.Hidden) != FileAttributes.Hidden)
						{
							list.Add(new DirectoryData(this, info.FullName, FSelectedPath));
						}
					}

					// リスト項目を昇順でソートする
					list.Sort(SortComparison);

					FSubItems = new List<DirectoryData>(list);
				}

				return FSubItems; 
			}
		}

		//-------------------------------------------------------------------------------------------
		// コンストラクタ
		// parentData   : 親データ（null のとき、ルートディレクトリつまりドライブである）
		// dirName      : ディレクトリ名（ルートディレクトリの場合、たとえば c:\）
		// selectedPath : コントロールの初期値
		public DirectoryData(DirectoryData parentData, string dirName, string selectedPath)
		{
			DriveInfo driveInfo = null;

			if (parentData == null)
				driveInfo = new DriveInfo(dirName);

			FSelectedPath = selectedPath;
			FDirectoryInfo = new DirectoryInfo(dirName);

			//------------------------
			if (parentData == null) // ルートディレクトリのとき
				FDirectoryIcon = GetFolderIcon(FDirectoryInfo.FullName);
			else
				FDirectoryIcon = FDefaultIcon; // ドライブ以外は取得済みのデフォルトのアイコンを使う
			//------------------------

			FFullName = FDirectoryInfo.FullName;
			FDirectoryName = FDirectoryInfo.Name;
			FSubItems = new List<DirectoryData>();

			if (driveInfo != null) // ドライブのとき
			{
				if (driveInfo.IsReady) // ドライブの準備がいいとき
				{
					var dirInfo = new DirectoryInfo(driveInfo.Name);
					this.AddDummyData(dirInfo);
				}
			} 
			else // ディレクトリのとき
			{
				this.AddDummyData(FDirectoryInfo);
			}

			// SelectedPath に対応して、ノードを展開またはリスト項目を選択する
			this.SetExpandedAndSelectedProperty(dirName);
		}

		//---------------------------------------------------------------------------------------------
		// 一般のフォルダ用デフォルトのアイコンを取得する static なコンストラクタ
		static DirectoryData()
		{
			string path = Environment.GetFolderPath(Environment.SpecialFolder.System);

			FDefaultIcon = GetFolderIcon(path);
		}

		//---------------------------------------------------------------------------------------------
		// ダミー用データのコンストラクタ
		private DirectoryData()
		{
		}

		//-------------------------------------------------------------------------------------------
		// SubItems プロパティ内で使う Sort メソッドのための Comparison<T> デリゲート
		private int SortComparison(DirectoryData x, DirectoryData y)
		{
			if (x == null)
			{
				if (y == null) // null 同士のときは等値
					return 0;
				else           // x が null、y が null でないとき、常に y のほうが大きい 
					return -1;
			}
			else
			{
				if (y == null) // x が null でなく、y が null のとき、常に x のほうが大きい 
					return 1;
				else           // x、y ともに null でないとき
					return x.DirectoryName.CompareTo(y.DirectoryName);
			}
		}

		//---------------------------------------------------------------------------------------------
		// SelectedPath プロパティに対応して、ノードを展開またはリスト項目を選択する
		private void SetExpandedAndSelectedProperty(string dirName)
		{
			if ((FSelectedPath == String.Empty) || (FSelectedPath == null))
				return;

			if (Directory.Exists(FSelectedPath) == false)
				return;

			if (String.Compare(FSelectedPath, dirName, true) == 0) // 完全に一致するとき
			{
				FIsSelected = true; // 現在の項目を選択状態にする
			}
			else if (FSelectedPath.StartsWith(dirName, StringComparison.CurrentCultureIgnoreCase)) // 部分的に一致するとき
			{
				FIsExpanded = true; // 現在のノードを展開する
			}
		}

		//---------------------------------------------------------------------------------------------
		// 空のサブディレクトリを追加する
		private void AddDummyData(DirectoryInfo info)
		{
			var subInfos = info.GetDirectories();

			if (subInfos.Length > 0) // サブ項目を持つとき
			{
				FHasSubItems = true;

				var data = new DirectoryData(); // 空のデータを追加
				FSubItems.Add(data);
			}
		}

		//---------------------------------------------------------------------------------------------
		// フォルダ用アイコンを取得する
		private static BitmapSource GetFolderIcon(string path)
		{
			BitmapSource icon = null;

			SHFILEINFO psf = new SHFILEINFO();
			SHGetFileInfo(path, 0, ref psf, Marshal.SizeOf(psf), SHGFI_ICON | SHGFI_SMALLICON);

			if (psf.hIcon != IntPtr.Zero)
				icon = Imaging.CreateBitmapSourceFromHIcon(psf.hIcon, Int32Rect.Empty, null);

			return icon;
		}

		//-------------------------------------------------------------------------------------------
		protected void OnPropertyChanged(string propertyName)
		{
			var handler = PropertyChanged;

			if (handler != null)
			{
				var e = new PropertyChangedEventArgs(propertyName);
				handler(this, e);
			}
		}

		//***********************************************************************************************
		private struct SHFILEINFO
		{
			public IntPtr hIcon;
			public int iIcon;
			public uint dwAttributes;
			[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 260)]
			public string szDisplayName;
			[MarshalAs(UnmanagedType.ByValTStr, SizeConst = 80)]
			public string szTypeName;
		}

		private const uint SHGFI_ICON = 0x100;
		private const uint SHGFI_LARGEICON = 0x0;
		private const uint SHGFI_SMALLICON = 0x1;

		[DllImport("shell32.dll")]
		private static extern int SHGetFileInfo(string pszPath, uint dwFileAttributes,
			ref SHFILEINFO psfi, int cbSizeFileInfo, uint uFlags);

	} // end of DirectoryData class

} // end of namespace
